"""Prime number calculations.

The functions "factorise" and "factors" are copied from the `pyprimes
library <https://github.com/uzumaxy/pyprimes/blob/master/src/pyprimes/factors.py>`_.
In mutwo the function :func:`pyprimes.primes` has been replaced by
:class:`primesieve.Iterator()`, which improves speed by a factor of 10.
"""

import typing

import primesieve  # type: ignore

__all__ = ("factorise", "factors", "is_prime")


class PrimeGenerator(object):
    """Generator that yields the rising series of primes starting from 2."""

    def __init__(self):
        self.iterator = primesieve.Iterator()

    def __next__(self) -> int:
        return self.iterator.next_prime()

    def __iter__(self) -> typing.Iterator:
        return self


def factorise(number_to_factorise: int) -> list[int]:
    """factorise(integer) -> [list of factors]

    :param number_to_factorise: The number which shall be factorised.
    :return: Returns a list of the (mostly) prime factors of integer n. For negative
        integers, -1 is included as a factor. If n is 0, 1 or -1, [n] is
        returned as the only factor. Otherwise all the factors will be prime.

    **Example:**

    >>> factorise(-693)
    [-1, 3, 3, 7, 11]
    >>> factorise(55614)
    [2, 3, 13, 23, 31]
    """
    result_list = []
    for prime_number, prime_number_count in factors(number_to_factorise):
        result_list.extend([prime_number] * prime_number_count)
    return result_list


def factors(number: int) -> typing.Generator:
    """Get factor generator

    :param number: The number from which to yield factors.

    Yields tuples of (factor, count) where each factor is unique and usually
    prime, and count is an integer 1 or larger.
    The factors are prime, except under the following circumstances: if the
    argument n is negative, -1 is included as a factor; if n is 0 or 1, it
    is given as the only factor. For all other integer n, all of the factors
    returned are prime.

    **Example:**

    >>> list(factors(3*7*7*7*11))
    [(3, 1), (7, 3), (11, 1)]
    """
    if number in (0, 1, -1):
        yield (number, 1)
        return
    elif number < 0:
        yield (-1, 1)
        number = -number
    assert number >= 2
    for prime_number in PrimeGenerator():
        if prime_number * prime_number > number:
            break
        count = 0
        while number % prime_number == 0:
            count += 1
            number //= prime_number
        if count:
            yield (prime_number, count)
    if number != 1:
        yield (number, 1)


def is_prime(number_to_test: int) -> bool:
    """Test if number is prime or not.

    :param number_to_test: The number which shall be tested.
    :return: True if number is prime and False if number isn't a Prime.

    (`has been copied from here
    <https://www.geeksforgeeks.org/python-program-to-check-
    whether-a-number-is-prime-or-not/>`_)
    """

    # Corner cases
    if number_to_test <= 1:
        return False
    if number_to_test <= 3:
        return True

    # This is checked so that we can skip
    # middle five numbers in below loop
    if number_to_test % 2 == 0 or number_to_test % 3 == 0:
        return False

    i = 5
    while i * i <= number_to_test:
        if number_to_test % i == 0 or number_to_test % (i + 2) == 0:
            return False
        i = i + 6

    return True



"""Utility functions."""

from . import configurations

from .decorators import *
from .exceptions import *
from .prime_factors import *
from .tools import *

from . import decorators, exceptions, prime_factors, tools

__all__ = tools.get_all(decorators, exceptions, prime_factors, tools)

# Force flat structure
del decorators, exceptions, prime_factors, tools



"""Standardization for transformations between parameters and simple events

Adds classes to allow transformations in two directions:

1. Extract data (e.g. mutwo parameters) from simple events
2. Convert data (e.g. mutwo parameters) to simple events
"""

import typing

from mutwo import core_constants
from mutwo import core_converters
from mutwo import core_events
from mutwo import core_utilities


__all__ = (
    "SimpleEventToAttribute",
    "MutwoParameterDict",
    "MutwoParameterDictToKeywordArgument",
    "MutwoParameterDictToDuration",
    "MutwoParameterDictToSimpleEvent",
    "UnknownObjectToObject",
)


class SimpleEventToAttribute(core_converters.abc.Converter):
    """Extract from a simple event an attribute.

    :param attribute_name: The name of the attribute which is fetched
        from a :class:`mutwo.core_events.SimpleEvent`.
    :param exception_value: This value is returned in case an `AttributeError`
        raises .
    """

    def __init__(self, attribute_name: str, exception_value: typing.Any):
        self._attribute_name = attribute_name
        self._exception_value = exception_value

    def convert(self, simple_event_to_convert: core_events.SimpleEvent) -> typing.Any:
        """Extract from a :class:`mutwo.core_events.SimpleEvent` an attribute.

        :param simple_event_to_convert: The :class:`mutwo.core_events.SimpleEvent`
            from which an attribute shall be extracted.
        :type simple_event_to_convert: mutwo.core_events.SimpleEvent

        **Example:**

        >>> from mutwo import core_converters
        >>> from mutwo import core_events
        >>> simple_event = core_events.SimpleEvent(duration=10)
        >>> simple_event_to_duration = core_converters.SimpleEventToAttribute(
                'duration', 0
            )
        >>> simple_event_to_duration.convert(simple_event)
        10
        >>> simple_event_to_pasta = core_converters.SimpleEventToAttribute(
                'pasta', 'spaghetti'
            )
        >>> simple_event_to_pasta.convert(simple_event)
        'spaghetti'
        >>> simple_event.pasta = 'tagliatelle'
        >>> simple_event_to_pasta.convert(simple_event)
        'tagliatelle'
        """

        return core_utilities.call_function_except_attribute_error(
            lambda simple_event: getattr(simple_event, self._attribute_name),
            simple_event_to_convert,
            self._exception_value,
        )


MutwoParameterDict = dict[str, core_constants.ParameterType]


class MutwoParameterDictToKeywordArgument(core_converters.abc.Converter):
    """Extract from a dict of mutwo parameters specific objects.

    :param mutwo_parameter_to_search_name: The parameter name which
        should be fetched from the MutwoParameterDict (if it exists).
    :type mutwo_parameter_to_search_name: str
    :param keyword: The keyword string to return. If no argument is given
        it will use the same value as :param:`mutwo_parameter_to_search_name`.
    :type keyword: typing.Optional[str]

    **Example:**

    >>> from mutwo import core_converters
    >>> from mutwo import music_parameters
    >>> mutwo_parameter_dict_to_keyword_argument = core_converters.MutwoParameterDictToKeywordArgument('pitch')
    >>> mutwo_parameter_dict_to_keyword_argument.convert(
        {'pitch': music_parameters.WesternPitch('c')}
    )
    ('pitch', music_parameters.WesternPitch(c4))
    """

    def __init__(
        self, mutwo_parameter_to_search_name: str, keyword: typing.Optional[str] = None
    ):
        if keyword is None:
            keyword = str(mutwo_parameter_to_search_name)
        self._mutwo_parameter_to_search_name = mutwo_parameter_to_search_name
        self._keyword = keyword

    def convert(
        self, mutwo_parameter_dict_to_convert: MutwoParameterDict
    ) -> typing.Optional[tuple[str, core_constants.ParameterType]]:
        try:
            return (
                self._keyword,
                mutwo_parameter_dict_to_convert[self._mutwo_parameter_to_search_name],
            )
        except KeyError:
            return None


class MutwoParameterDictToDuration(MutwoParameterDictToKeywordArgument):
    """Extract from a dict of mutwo parameters the duration.

    :param duration_to_search_name: The name of the duration which shall
        be searched for in the :const:`MutwoParameterDict`. If `None` the
        value of the global constants
        :const:`mutwo.core_converters.configurations.DEFAULT_DURATION_TO_SEARCH_NAME`
        will be used. Default to `None`.
    :type duration_to_search_name: typing.Optional[str]
    :param duration_keyword_name: The name of the duration keyword for the
        event. If `None` the value of the global constants
        :const:`mutwo.core_converters.configurations.DEFAULT_DURATION_KEYWORD_NAME`
        will be used. Default to `None`.
    :type duration_keyword_name: typing.Optional[str]
        :const:`mutwo.core_converters.configurations.DEFAULT_DURATION_KEYWORD_NAME`.
    """

    def __init__(
        self,
        duration_to_search_name: typing.Optional[str] = None,
        duration_keyword_name: typing.Optional[str] = None,
    ):
        if duration_to_search_name is None:
            duration_to_search_name = (
                core_converters.configurations.DEFAULT_DURATION_TO_SEARCH_NAME
            )
        if duration_keyword_name is None:
            duration_keyword_name = (
                core_converters.configurations.DEFAULT_DURATION_KEYWORD_NAME
            )
        assert isinstance(duration_to_search_name, str)
        assert isinstance(duration_keyword_name, str)
        super().__init__(duration_to_search_name, duration_keyword_name)


class MutwoParameterDictToSimpleEvent(core_converters.abc.Converter):
    """Convert a dict of mutwo parameters to a :class:`mutwo.core_events.SimpleEvent`

    :param mutwo_parameter_dict_to_keyword_argument_sequence: A sequence of
        :class:`MutwoParameterDictToKeywordArgument`. If set to `None`
        a sequence with :class:`MutwoParameterDictToDuration` will be created.
        Default to `None`.
    :type mutwo_parameter_dict_to_keyword_argument_sequence: typing.Optional[typing.Sequence[MutwoParameterDictToKeywordArgument]]
    :param simple_event_class: Default to :class:`mutwo.core_events.SimpleEvent`.
    :type simple_event_class: typing.Type[core_events.SimpleEvent]
    """

    def __init__(
        self,
        mutwo_parameter_dict_to_keyword_argument_sequence: typing.Optional[
            typing.Sequence[MutwoParameterDictToKeywordArgument]
        ] = None,
        simple_event_class: typing.Type[
            core_events.SimpleEvent
        ] = core_events.SimpleEvent,
    ):
        if mutwo_parameter_dict_to_keyword_argument_sequence is None:
            mutwo_parameter_dict_to_keyword_argument_sequence = (
                MutwoParameterDictToDuration(),
            )
        self._mutwo_parameter_dict_to_keyword_argument_sequence = (
            mutwo_parameter_dict_to_keyword_argument_sequence
        )
        self._simple_event_class = simple_event_class

    def convert(
        self, mutwo_parameter_dict_to_convert: MutwoParameterDict
    ) -> core_events.SimpleEvent:
        keyword_argument_dict = {}
        for (
            mutwo_parameter_dict_to_keyword_argument
        ) in self._mutwo_parameter_dict_to_keyword_argument_sequence:
            keyword_argument_or_none = mutwo_parameter_dict_to_keyword_argument.convert(
                mutwo_parameter_dict_to_convert
            )
            if keyword_argument_or_none:
                keyword, argument = keyword_argument_or_none
                keyword_argument_dict.update({keyword: argument})
        return self._simple_event_class(**keyword_argument_dict)


T = typing.TypeVar("T")


class UnknownObjectToObject(core_converters.abc.Converter, typing.Generic[T]):
    """Helper to simplify standardisation of syntactic sugar.

    :param type_tuple_to_callable_dict: Define which types are converted by
        which methods.

    **Example:**

    >>> from mutwo impot core_converters
    >>> anything_to_string = core_converters.UnknownObjectToObject[str](
    >>>     (
    >>>         ((float, int, list), str),
    >>>         ((tuple,), lambda t: str(len(t))),
    >>>         ([], lambda _: "..."),
    >>>     )
    >>> )
    >>> anything_to_string.convert(100)
    "100"
    >>> anything_to_string.convert(7.32)
    "7.32"
    >>> anything_to_string.convert((1, 2, 3))
    "3"
    >>> anything_to_string.convert(b'')
    "..."
    """

    def __init__(
        self,
        type_tuple_and_callable_tuple: tuple[tuple[typing.Type, ...], typing.Callable],
    ):
        self._type_tuple_and_callable_tuple = type_tuple_and_callable_tuple

    def convert(self, unknown_object_to_convert: typing.Any) -> T:
        # XXX: This may break in the future, because it is an implementation
        # detail.
        if isinstance(unknown_object_to_convert, typing.get_args(self.__orig_class__)):
            return unknown_object_to_convert
        for type_tuple, callable_object in self._type_tuple_and_callable_tuple:
            if type_tuple:
                if isinstance(unknown_object_to_convert, type_tuple):
                    return callable_object(unknown_object_to_convert)
            else:
                return callable_object(unknown_object_to_convert)

        raise NotImplementedError(
            f"No conversion routine defined for object '{unknown_object_to_convert}'"
            f" of type '{type(unknown_object_to_convert)}'."
        )



"""Configurations which are shared for all parameter classes in :mod:`mutwo.core_parameters`."""

ROUND_DURATION_TO_N_DIGITS = 10
"""Set floating point precision for the `duration_in_floats` property of all
:class:`~mutwo.core_parameters.abc.Duration` classes in the :mod:`mutwo.core_parameters`
module.

When returning the `duration_in_floats` property all mentioned events will
round their actual duration if the duration type is float. This
behaviour has been added with version 0.28.1 to avoid floating
point rounding errors which could occur in all duration related
methods of the different event classes (as it can happen in
for instance the :func:`mutwo.core_events.abc.ComplexEvent.squash_in`
method or the :func:`mutwo.core_events.abc.Event.cut_off`
method)."""



"""Abstract base classes for different parameters.

This module defines the public API of parameters.
Most other mutwo classes rely on this API. This means
when someone creates a new class inheriting from any of the
abstract parameter classes which are defined in this module,
she or he can make use of all other mutwo modules with this
newly created parameter class.
"""

from __future__ import annotations

import abc
import functools
import operator
import typing

try:
    import quicktions as fractions
except ImportError:
    import fractions

from mutwo import core_constants
from mutwo import core_events
from mutwo import core_parameters
from mutwo import core_utilities

__all__ = (
    "SingleValueParameter",
    "SingleNumberParameter",
    "ParameterWithEnvelope",
    "Duration",
)


class ParameterWithEnvelope(abc.ABC):
    """Abstract base class for all parameters with an envelope."""

    def __init__(self, envelope: core_events.RelativeEnvelope):
        self.envelope = envelope

    @property
    def envelope(self) -> core_events.RelativeEnvelope:
        return self._envelope

    @envelope.setter
    def envelope(self, new_envelope: typing.Any):
        if not isinstance(new_envelope, core_events.RelativeEnvelope):
            raise TypeError(
                f"Found illegal object '{new_envelope}' of not "
                f"supported type '{type(new_envelope)}'. "
                f"Only instances of '{core_events.RelativeEnvelope}'"
                " are allowed!"
            )
        self._envelope = new_envelope

    def resolve_envelope(
        self,
        duration: core_constants.DurationType,
        # XXX: We can't directly set the default attribute value,
        # but we have to do it with `None` and resolve it later,
        # because otherwise we will get a circular import
        # (core_parameters need to be imported before core_events,
        #  because we need core_parameters.Duration in core_events).
        resolve_envelope_class: typing.Optional[type[core_events.Envelope]] = None,
    ) -> core_events.Envelope:
        if resolve_envelope_class is None:
            resolve_envelope_class = core_events.Envelope
        return self.envelope.resolve(duration, self, resolve_envelope_class)


class SingleValueParameter(abc.ABC):
    """Abstract base class for all parameters which are defined by one value.

    Classes which inherit from this base class have
    to provide an additional keyword argument `value_name`.
    Furthermore they can provide the optional keyword argument
    `value_return_type`.

    **Example:**

    >>> from mutwo import core_parameters
    >>> class Color(
            core_parameters.abc.SingleValueParameter,
            value_name="color",
            value_return_type=str
        ):
            def __init__(self, color: str):
                self._color = color
            @property
            def color(self) -> str:
                return self._color
    >>> red = Color('red')
    >>> red.color
    'red'
    >>> orange = Color('orange')
    >>> red2 = Color('red')
    >>> red == orange
    False
    >>> red == red2
    True
    """

    def __init_subclass__(
        cls, value_name: str = "", value_return_type: typing.Type = typing.Any
    ):
        # We will only add an abstract method with "value_name" in two
        # cases for the following reasons
        #
        #   1. If our value_name is empty we shouldn't add any new method.
        #      With this we can prevent that a new abstract property is created
        #      when no value_name is provided and the default value_name isn't
        #      empty. This would be a problem if for instance we would have the
        #      following inheritance structure:
        #
        #      SingleValueParameter ->
        #
        #           Pitch(SingleValueParameter, value_name="frequency") ->
        #
        #               MidiPitch(Pitch)
        #
        #       The last class shouldn't get any new abstract property (with
        #       a not-empty default name).
        #
        #   2. If the class already defines a method with the specific
        #      value_name it shouldn't be overridden (because the user already
        #      ensured there is a property). Overriding this manually defined property
        #      would lead to unexpected and undesired results.
        if value_name:
            if not hasattr(cls, value_name):

                @abc.abstractmethod
                def abstract_method(_) -> value_return_type:
                    raise NotImplementedError

                setattr(cls, value_name, property(abstract_method))

            if hasattr(cls, "value_name"):
                raise core_utilities.AlreadyDefinedValueNameError(cls)

            setattr(cls, "value_name", property(lambda _: value_name))

    def __str__(self) -> str:
        return (
            f"{type(self).__name__}"
            f"({self.value_name} = {getattr(self, self.value_name)})"  # type: ignore
        )

    def __eq__(self, other: typing.Any) -> bool:
        try:
            return getattr(self, self.value_name) == getattr(other, self.value_name)  # type: ignore
        except AttributeError:
            return False


@functools.total_ordering
class SingleNumberParameter(SingleValueParameter):
    """Abstract base class for all parameters which are defined by one number.

    Classes which inherit from this base class have to override
    the same methods and properties as one have to override
    when inheriting from :class:`SingleValueParameter`.

    Furthermore the property `digit_to_round_to_count`
    can be overridden. This should return an integer or `None`.
    If it returns an integer it will first round two numbers
    before comparing them with the `==` or `<` or `<=` or `>`
    or `>=` operators.
    The default implementation always returns `None.

    **Example:**

    >>> from mutwo import core_parameters
    >>> class Speed(
            core_parameters.abc.SingleNumberParameter,
            value_name="meter_per_seconds",
            value_return_type=float
        ):
            def __init__(self, meter_per_seconds: float):
                self._meter_per_seconds = meter_per_seconds
            @property
            def meter_per_seconds(self) -> float:
                return self._meter_per_seconds
    >>> light_speed = Speed(299792458)
    >>> sound_speed = Speed(343)
    >>> light_speed > sound_speed
    True
    """

    direct_comparison_type_tuple = tuple([])

    @property
    def digit_to_round_to_count(self) -> typing.Optional[int]:
        return None

    def _prepare_value_pair_for_comparison(
        self, value_pair: tuple[core_constants.Real, core_constants.Real]
    ) -> tuple[core_constants.Real, core_constants.Real]:
        return tuple(
            core_utilities.round_floats(value, self.digit_to_round_to_count)
            if self.digit_to_round_to_count
            else value
            for value in value_pair
        )

    def _compare(
        self,
        other: typing.Any,
        compare: typing.Callable[[core_constants.Real, core_constants.Real], bool],
        raise_exception: bool,
    ):
        """Compare itself with other object"""

        try:
            value_pair = (
                getattr(self, self.value_name),  # type: ignore
                other
                if isinstance(other, self.direct_comparison_type_tuple)
                else getattr(other, self.value_name),  # type: ignore
            )
        except AttributeError:
            if raise_exception:
                raise TypeError(
                    f"Can't compare object '{self}' of type '{type(self)}' with"
                    f" object '{other}' of type '{type(other)}'!"
                )
            return False

        value0, value1 = self._prepare_value_pair_for_comparison(value_pair)
        return compare(value0, value1)

    def __float__(self) -> float:
        return float(getattr(self, self.value_name))  # type: ignore

    def __int__(self) -> int:
        return int(float(self))

    def __eq__(self, other: typing.Any) -> bool:
        return self._compare(other, lambda value0, value1: value0 == value1, False)

    def __lt__(self, other: typing.Any) -> bool:
        return self._compare(other, lambda value0, value1: value0 < value1, True)


class Duration(
    SingleNumberParameter, value_name="duration", value_return_type="fractions.Fraction"
):
    """Abstract base class for any duration.

    If the user wants to define a Duration class, the abstract
    property :attr:`duration` has to be overridden.

    The attribute :attr:`duration` is stored in unit `beats`.
    """

    direct_comparison_type_tuple = (float, int, fractions.Fraction)

    def _math_operation(
        self, other: DurationOrReal, operation: typing.Callable[[float, float], float]
    ) -> Duration:
        if isinstance(other, core_constants.Real.__args__):
            other = core_parameters.DirectDuration(float(other))
        self.duration = fractions.Fraction(operation(self.duration, other.duration))
        return self

    @core_utilities.add_copy_option
    def add(self, other: DurationOrReal) -> Duration:
        return self._math_operation(other, operator.add)

    @core_utilities.add_copy_option
    def subtract(self, other: DurationOrReal) -> Duration:
        return self._math_operation(other, operator.sub)

    @core_utilities.add_copy_option
    def multiply(self, other: DurationOrReal) -> Duration:
        return self._math_operation(other, operator.mul)

    @core_utilities.add_copy_option
    def divide(self, other: DurationOrReal) -> Duration:
        return self._math_operation(other, operator.truediv)

    def __add__(self, other: DurationOrReal) -> Duration:
        return self.add(other, mutate=False)

    def __sub__(self, other: DurationOrReal) -> Duration:
        return self.subtract(other, mutate=False)

    def __mul__(self, other: DurationOrReal) -> Duration:
        return self.multiply(other, mutate=False)

    def __truediv__(self, other: DurationOrReal) -> Duration:
        return self.divide(other, mutate=False)

    def __float__(self) -> float:
        return core_utilities.round_floats(
            float(self.duration),
            core_parameters.configurations.ROUND_DURATION_TO_N_DIGITS,
        )

    @property
    def duration_in_floats(self) -> float:
        return float(self)

    @property
    def duration(self) -> fractions.Fraction:
        ...

    @duration.setter
    @abc.abstractmethod
    def duration(self, duration: fractions.Fraction):
        ...


DurationOrReal = typing.Union[Duration, core_constants.Real]



"""Convert data from and to mutwo."""

from . import abc
from . import configurations

from .parsers import *
from .tempos import *

from . import parsers, tempos

from mutwo import core_utilities

__all__ = core_utilities.get_all(parsers, tempos)

# Force flat structure
del core_utilities, parsers, tempos



"""Submodule for the parameter tempo.
"""

import typing

from mutwo import core_constants
from mutwo import core_utilities

__all__ = ("TempoPoint",)

TempoInBeatsPerMinute = float
TempoRangeInBeatsPerMinute = tuple[TempoInBeatsPerMinute, TempoInBeatsPerMinute]
TempoOrTempoRangeInBeatsPerMinute = typing.Union[
    TempoInBeatsPerMinute, TempoRangeInBeatsPerMinute
]


class TempoPoint(object):
    """Represent the active tempo at a specific moment in time.

    :param tempo_or_tempo_range_in_beats_per_minute: Specify a tempo in
        `beats per minute <https://en.wikipedia.org/wiki/Tempo#Measurement>`_.
        Tempo can also be a tempo range where the first value indicates a minimal
        tempo and the second value the maximum tempo. If the user specifies a
        range :mod:`mutwo` will use the minimal tempo in internal calculations.
    :param reference: The reference with which the tempo will be multiplied.
        In terms of Western notation a reference = 1 will be a 1/4 beat, a
        reference of 2 will be a 1/2 beat, etc. Default to 1.
    :param textual_indication: Sometimes it is desired to specify an extra
        text indication how fast or slow the music should be (for instance
        "Adagio" in Western music). Default to `None`.

    **Example:**

    >>> from mutwo import core_events
    >>> from mutwo import core_parameters
    >>> tempo_envelope = core_events.TempoEnvelope([
    >>>     [0, core_parameters.TempoPoint(60, reference=2)]
    >>> ])
    """

    def __init__(
        self,
        tempo_or_tempo_range_in_beats_per_minute: TempoOrTempoRangeInBeatsPerMinute,
        reference: core_constants.Real = 1,
        textual_indication: typing.Optional[str] = None,
    ):
        self.tempo_or_tempo_range_in_beats_per_minute = (
            tempo_or_tempo_range_in_beats_per_minute
        )
        self.reference = reference
        self.textual_indication = textual_indication

    def __repr__(self) -> str:
        return "{}(BPM = {}, reference = {})".format(
            type(self).__name__, self.tempo_in_beats_per_minute, self.reference
        )

    def __eq__(self, other: object) -> bool:
        attribute_to_compare_tuple = (
            "tempo_in_beats_per_minute",
            "reference",
            "textual_indication",
        )
        return core_utilities.test_if_objects_are_equal_by_parameter_tuple(
            self, other, attribute_to_compare_tuple
        )

    @property
    def tempo_in_beats_per_minute(self) -> TempoInBeatsPerMinute:
        """Get tempo in `beats per minute <https://en.wikipedia.org/wiki/Tempo#Measurement>`_

        If :attr:`tempo_or_tempo_range_in_beats_per_minute` is a range
        mutwo will return the minimal tempo.
        """

        if isinstance(self.tempo_or_tempo_range_in_beats_per_minute, tuple):
            return self.tempo_or_tempo_range_in_beats_per_minute[0]
        else:
            return self.tempo_or_tempo_range_in_beats_per_minute

    @property
    def absolute_tempo_in_beats_per_minute(self) -> float:
        """Get absolute tempo in `beats per minute <https://en.wikipedia.org/wiki/Tempo#Measurement>`_

        The absolute tempo takes the :attr:`reference` of the :class:`TempoPoint`
        into account.
        """

        return self.tempo_in_beats_per_minute * self.reference



"""Time-based Event abstractions.

Event objects can be understood as the core objects
of the :mod:`mutwo` framework. They all own a :attr:`duration`
attribute (which can be any number). Further more complex
Event classes with more relevant attributes can be generated
through inheriting from basic classes. :mod:`mutwo` already offers
support for several more complex representations (for instance
:class:`mutwo.music_events.NoteLike`).
The most often used classes may be:
    - :class:`mutwo.core_events.SimpleEvent`
    - :class:`mutwo.core_events.SequentialEvent`
    - :class:`mutwo.core_events.SimultaneousEvent`
"""

from . import configurations
from . import abc

from .basic import *
from .envelopes import *

from . import basic, envelopes

from mutwo import core_utilities

__all__ = core_utilities.get_all(basic, envelopes)

# Force flat structure
del basic, core_utilities, envelopes
